# -*- coding: UTF-8 -*-
from __future__ import print_function

import os
import os.path
import errno
from glob import glob
from unidecode import unidecode
import re
import requests

DICTS_URL = 'http://bfontaine.net/crosswords/dicts'
DICTS_PATH = os.path.join(os.path.expanduser('~'), '.crosswords', 'dicts')

DEFAULT = 'english'


def sanitize_word(word):
    """
    sanitize a word by removing its accents, special characters, etc
    """
    # use an unicode string for `unidecode`
    if type(word) == str:
        try:
            word = word.decode()
        except AttributeError:
            pass  # Python3

    # remove trailing spaces
    word = word.strip()

    # remove stuff between parentheses
    word = re.sub(r'\([^)]*\)', '', word)

    # remove accents, hyphens & other special chars
    word = re.sub(r'[ "\'-;.]+', '', unidecode(word))

    # only lowercase
    return word.lower()


def init_storage():
    """
    Initialize the local dictionnaries cache in ~/.crosswords/dicts.
    """
    # http://stackoverflow.com/a/600612/735926
    try:
        os.makedirs(DICTS_PATH)
    except OSError as ex:
        if ex.errno != errno.EEXIST or not os.path.isdir(DICTS_PATH):
            raise


def local_list(timestamps=True):
    """
    Return a list of the locally available dictionnaries. Each element is a
    tuple of the dictionnary name and its last modification date as a
    timestamp.
    """
    init_storage()
    lst = []
    for d in glob(os.path.join(DICTS_PATH, '*.txt')):
        name = d.split('/')[-1].replace('.txt', '')
        lst.append((name, os.path.getmtime(d)) if timestamps else name)

    return lst


def remote_list(timestamps=True):
    """
    Return a list of the remotely available dictionnaries. Each element is a
    tuple of the dictionnary name and its last modification date as a
    timestamp.
    """
    r = requests.get(DICTS_URL)
    lst = []
    for f in r.text.split('\n'):
        if not f:
            continue
        name, date = f.split(':')
        name = name.replace('.txt', '')
        lst.append((name, int(date)) if timestamps else name)

    return lst


def update(verbose=False):
    """
    Update local dictionnaries by downloading the latest version from the
    server, if there's one.
    """
    local = local_list()
    remote = dict(remote_list())

    updated = False

    for name, date in local:
        if name in remote and remote[name] > date:
            updated = True
            if verbose:
                print("Updating '%s'..." % name)
            download(name)

    if not updated and verbose:
        print("Nothing to update.")


def download(name, verbose=False):
    """
    Download a dictionnary from the remote server into the local cache. Return
    the number of new words or -1 on error.
    """
    init_storage()

    if verbose:
        print("Downloading '%s'..." % name)
    r = requests.get('%s/%s.txt' % (DICTS_URL, name))

    if r.status_code != 200:
        return -1

    wcount = 0

    with open(os.path.join(DICTS_PATH, '%s.txt' % name), 'w') as f:
        for word in r.iter_lines():
            wcount += 1
            f.write('%s\n' % sanitize_word(word))

    if verbose:
        print("%d words added." % wcount)
    return wcount


def filepath(language):
    """return a file path for a given language"""
    return os.path.join(DICTS_PATH, '%s.txt' % language)


def exists(language):
    """test if a dictionnary for the given language exists locally"""
    return os.path.exists(filepath(language))



# -*- coding: UTF-8 -*-
from __future__ import print_function

import sys
import crosswords
from crosswords.words import get_matches
import crosswords.dictionnaries as dicts


def print_help_and_exit(exe='crosswords'):
    print("Usage:\n\t%s [-l <language>] <word>\n" % exe)
    print("""\
Additionnally, some words have a special meaning:
  install:     install a new language
  list:        list local languages
  remote_list: list installable languages
  update:      update local languages""")
    sys.exit(1)


def find_word(word, dictionnary):
    matches = get_matches(word, dictionnary)
    c = len(matches)

    if not matches:
        print("No word found.")
    else:
        print("%d possibilit%s:" % (c, 'y' if c == 1 else 'ies'))
        for match in matches:
            print('- %s' % match)


def print_languages_and_exit(lst, status=1, header=True):
    """print a list of languages and exit"""
    if header:
        print("Available languages:")
    for lg in lst:
        print("- %s" % lg)
    sys.exit(status)


def extract_opts(**opts):
    """
    Small utility to extract a set of one-char options from sys.argv.
    """
    values = {}
    for opt, init in opts.items():
        try:
            idx = sys.argv.index('-%s' % opt)
        except ValueError:
            continue
        if idx+1 < len(sys.argv):
            opts[opt] = sys.argv.pop(idx+1)

        sys.argv.pop(idx)

    return opts


def run():
    opts = extract_opts(l=dicts.DEFAULT)

    argc = len(sys.argv)
    if argc < 2:
        print_help_and_exit(sys.argv[0])

    word = sys.argv[1]

    if word == 'update':
        dicts.update(verbose=True)

    elif word == 'install':
        lst = dicts.remote_list(timestamps=False)
        if argc < 3:
            print("Usage:\n\t%s install <language>" % sys.argv[0])
            print_languages_and_exit(lst)

        lg = sys.argv[2].strip().lower()
        if lg not in lst:
            print("The language '%s' is not available." % lg)
            print_languages_and_exit(lst)
        dicts.download(lg)

    elif word == 'list':
        print("Locally available languages:")
        print_languages_and_exit(dicts.local_list(timestamps=False),
                                 status=0, header=False)
        print("Use 'remote_list' to list remotely available languages.")

    elif word == 'remote_list':
        print("Remotely available languages:")
        print_languages_and_exit(dicts.remote_list(timestamps=False),
                                 status=0, header=False)
        print("Use 'install <language>' to download and install a language.")

    elif word == '--version':
        print("Crosswords v%s" % crosswords.__version__)

    else:
        find_word(word, opts['l'])



# -*- coding: UTF-8 -*-

import re
import crosswords.dictionnaries as dicts


def compile_pattern(word):
    """
    take a word pattern and return a Python regexp. A word pattern is a word
    with unknown letters replaced by a '?', e.g. 'di?ti??nar?'.
    """
    return re.compile(r'^%s$' % re.sub(r'\?', '[a-z]',
                                       dicts.sanitize_word(word)))


def get_matches(pattern, language, max_count=8):
    """
    take a word pattern or a Python regexp and a language name, and return a
    list of all matching words.
    """
    if str(pattern) == pattern:
        pattern = compile_pattern(pattern)

    results = []

    if not dicts.exists(language):
        print("The language '%s' is not available locally." % language)
        return []

    with open(dicts.filepath(language), 'r') as f:
        for word in f:
            if max_count <= 0:
                break
            w = word.strip()
            if pattern.match(w):
                results.append(w)
                max_count -= 1

    return results
